/*
	detuned.cpp
	Copyright Risto Holopainen 2021

	Generator of the composition Auto-detune
	output: three separate soundfiles intended to be mixed or played simultaneously
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "detuned.h"
#include "leakmap.h"
#include "../util/delay_t.h"
#include "../util/azcr.h"
#include <vector>
#include <algorithm>	// sort
#include <cstdio>
#include <cmath> 
#include <ctime>
#include <sndfile.h>


using namespace detune;
using std::vector;



void stats(const vector<double> &x, const char *info)
{
	printf("STATS. ");
	// min,max,mean
	double min, max, s=0;
	min = max = x[0];
	const int N = x.size();
	
	for(int i=0; i<N; i++)
		{
		s += x[i];
		if(x[i] > max) max = x[i];
		if(x[i] < min) min = x[i];
		}
	double avg = s / N;
	vector <double> y = x;
	sort(y.begin(), y.end());
	double median;
	if(N & 1)	// odd
		median = y[(N-1)/2];
	else
		median = .5 * (y[(N-1)/2] + y[(N+1)/2]);

	printf("%s: min-max %.3f--%.3f,\t avg %.4f, \t median %.3f\n", 
		info, min, max, avg, median);
}


double rate_wig(double u, double v)
{
	double x = (u < 0) ? 0 : u; 
	double y = (v < 0) ? 0 : v;
	if(x > 1) x = 1;
	if(y > 1) y = 1;

	double A = 200 + 6000*x - 500*y;
	if(A < 10)	// negative really
		A = 20 - A;
	if(A > 16000)
		A = 15000;
	return A;
}


template <typename id>
void init_delay(delay<id> &del, id val, int N)
{
	for(int i=0; i<N; i++)
		del.z(val);
}



int main(int argc, char **argv)
{
	
	if(argc > 2)
		{
		printf("%s [modulation>0]\n\n", argv[0]);	
		exit(1);
		}
	
	double c0 = 20;
	if(argc == 2)
		c0 = atof(argv[1]);
	
	const int sr = 48000;
	detuned deta(sr, c0, detuned::SAW), detb(sr, c0, detuned::PINK);
	detuned detc(sr, c0, detuned::PRIME);
	double x, y, z, w1 = 0, w2 = 0;

	SNDFILE *ouf, *wig, *zout;
	SF_INFO head, headw, headz;
	head.samplerate = sr;
	head.channels = 2;
	head.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	headw = headz = head;
	headz.channels = 1;
	ouf = sf_open("detuned_AB.wav", SFM_WRITE, &head);
	wig = sf_open("detuned_wig.wav", SFM_WRITE, &headw);
	zout = sf_open("detuned_zout.wav", SFM_WRITE, &headz);

	const int L = 1 << 9;
	double buf[2*L], FF = 40, GG = 55,  HH = 700;
	double bufwig[2*L], bufz[L];
	
	const int TotalTime = 275 * sr / L;
	int ExtraTime = 0;
	
	time_t tt = time(0);
	parameters pp, qq, rr;
	pp.p1 = 175; pp.p2 = 178; 
	qq = pp;
	deta.set_periods(pp);
	detc.set_periods(pp);
	
	delay <double> Delf(5, 10.0);
	int deltime_p = 0.81*sr/L, deltime_q = 0.97*sr/L;
	
	delay <parameters> Delp(deltime_p, pp), Delq(deltime_q, pp);
	azcr ZXL(25), ZXR(24), ZXc(31);
	double rate_1, rate_2, rate_3;
	int MM = 21000, NN = 17500;		// duration till next update
	
	leakage::counter LC1, LC2;
	deque<double> wiggle1, wiggle2;
	detune::unwiggler unwig1(sr), unwig2(sr);	
	vector<double> zv1, zv2, zv3; 
	
	for(int i=0, mm=0, nn=0; i<TotalTime+ExtraTime; i++)
		{
		qq = Delp.z(pp);
		rr = Delq.z(pp);
		detb.set_periods(pp);
		deta.set_periods(qq);
		detc.set_periods(rr);
		
		for(int j=0, k=0; j<L; j++, k += 2)
			{
			x = deta.gen2();
			y = detb.gen2();
			z = detc.gen2();
			rate_1 = ZXL.rate(x);
			rate_2 = ZXR.rate(y);
			rate_3 = ZXc.rate(w1 + w2);
			
			if(j%64 == 32)
				{			// collect zcr-stats
				zv1.push_back(rate_1);	
				zv2.push_back(rate_2);
				zv3.push_back(rate_3);
				}
			
			if(++mm >= MM)
				{
				mm = 0;
				update_p2(pp, rate_3);
				LC1 = leakage::n_iterations(fabs(z));
				MM = 0.05*sr + 4000*LC1.n;
				wiggle1 = leakage::expandeq(LC1.dq);
				unwig1.set_g(LC1.x);
				unwig1.set_a(rate_wig(rate_2, rate_3));
				if(LC1.n > 300)
					printf("time %5d -- leakage_1 %d iterations\n", i, LC1.n);
				++ExtraTime;
				deta.set_quantize(1 + LC1.n%8);
				detb.set_quantize(1 + LC2.n%8);
				}
			if(++nn >= NN)
				{
				nn = 0;
				FF = update_f2(rate_2);
				GG = update_f2(rate_1);
				HH = 4.0*Delf.z(FF - GG) + 100.0;
				deta.set_frq(FF);
				detb.set_frq(GG);
				detc.set_frq(HH);
				detc.set_quantize(LC1.n%6 + LC2.n%6 + 1);
				LC2 = leakage::n_iterations( .5*(fabs(x) + fabs(y)) );
				NN = 0.05*sr + 4000*LC2.n;
				wiggle2 = leakage::expandeq(LC2.dq);
				unwig2.set_g(LC2.x);
				unwig2.set_a(rate_wig(rate_1, rate_3));

				if(LC2.n > 300)
					printf("time %5d -- leakage_2 %d iterations\n", i, LC2.n);
				ExtraTime += 2;
				}

			w1 = unwig1.unwiggle(wiggle1);
			w2 = unwig2.unwiggle(wiggle2);
			
			buf[k  ] = .85*x;
			buf[k+1] = .85*y;
			bufwig[k  ] = .85*w1;
			bufwig[k+1] = .85*w2;
			bufz[j] = .8*z;
			}
		sf_write_double(ouf, buf, 2*L);
		sf_write_double(wig, bufwig, 2*L);
		sf_write_double(zout, bufz, L);
		
		if(ExtraTime > 525)
			ExtraTime = 500;
		}
	
	sf_close(ouf);
	sf_close(wig);
	sf_close(zout);

	stats(zv1, "zcr 1");
	stats(zv2, "zcr 2");
	stats(zv3, "zcr 3");
	printf("ExtraTime served: %d = %.2f sec.\n", 
		ExtraTime, (float)ExtraTime*L/sr);

	if(tt % 3 == 1)
		printf("stay detuned!\n\n");

	return 0;
}
