
/*
	quadmap.h
	<c> Risto Holopainen 2021

	contains quad oscillator + quadratic map


	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    
*/


#ifndef QUADMAP_H
#define QUADMAP_H


/*
	Oscillator after Martin Vicanek : A new quatrature oscillator
*/

class quadosc
{
	public:
		quadosc();
		void frq(double w);
		void gen(double &p, double &q);

	private:
		double u, v;
		double k1, k2;
		double tmp;
};

quadosc::quadosc()
{
	u = 1.0;
	v = 0.0;
	tmp = 0;
}


void quadosc::frq(double w)
{
	k1 = tan(0.5*w);
	k2 = 2*k1/(1 + k1*k1);
}


void quadosc::gen(double &p, double &q)
{
	tmp = u - k1*v;
	v += k2*tmp;
	u = tmp - k1*v;
	p = u;
	q = v;
}




/*	
	2D Sprott map
	Ref: eq 2-15 in:
	Zeraoulia & Sprott (2010). 2-D Quadratic Maps and 3-D ODE Systems. World Scientific.

	parameters --> type, ranges

	A, 	B
	0.7, -1.0 explode
	0.6, -0.9 (chaos)	-0.96 .. +1.64
	0.6, -1.0 (c)		-1.25 .. +1.61
	0.6, -1.1 explode
	0.5, -0.9 period 15
	0.5, -1.0 (c)		-1.01 .. +1.51
	0.5, -1.1 (c)		-1.24 .. +1.51
	0.5	-1.2 explode
	0.55, -1.05 (c)		-1.22 .. + 1.57
*/

double norm_range(double u)
{
// u: - 1.25 .. 1.75 --> -1..+1
	double v = u - 0.25;
	return .65*v;
}


class qmap
{
	public:
		qmap();
		void param(double p, double q);	/* p,q: 0...1 */
		void set_ab(double a, double b);
		void iterate(double &x, double &y);
		void init(double x1, double y1) {x=x1; y=y1;};

	private:
		double A, B;
		double x, y;
};


qmap::qmap()
{
	A = +0.6;
	B = -1.0;
}

void qmap::param(double p, double q)
{
	// p, q in [0,1]
	A =  0.5 + 0.1*p;
	B = -1.1 + 0.1*(p+q);

	//printf("p,q: %.3f, %.3f \t -> \t A,B: %.3f, %.3f\n",
	//	p, q, A, B);
}

void qmap::set_ab(double a, double b)
{
	A = a;
	B = b;
}

void qmap::iterate(double &u, double &v)
{
	x = 1 + A*u + B*v*v;
	y = u;
	u = x;
	v = y;
}


#endif

