/*
	demolir
	Copyright Risto Holopainen 2021

	generator of the piece Bâtir & démolir

	Run batir.cpp first to generate its output soundfile, which
	is then read by demolir.cpp. The composition is the output of this program.


	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/


#include "binop.h"
#include "../util/delay_t.h"
#include "../util/diff.h"
#include <sndfile.h>


class transient
{
	// detector of transients
	public:
		transient();
		double detect(double);

	private:
		diff *df;
		double a;
		double y;
};

transient::transient()
{
	df = new diff(9);
	a = 0;
	y = 0;
}


double transient::detect(double x)
{
	a = x*x;
	y = .5*(a + y);
	return df->filter(y);
}


using namespace binopsic;

class merge
{
	public:
		merge(int sr);
		~merge();
		void combine(const double *ibuf, const double *dbuf, double *obuf, int sz);

	private:
		void regime_change();
		int sr;
		delay<double> *del_a;
		delay<double> *del_b;
		delay<double> *del_c;
		delay<double> *del_d;
		bin_op *bop, *bup;
		int k, m;
		double x, y, xd, yd;
		double x1, y1;
		double u, v;
		double L, R;
		double fbx, fby;
		transient tr;
		double A, sa;
		long int t;
};


merge::merge(int fs)
{
	sr = fs;

	del_a = new delay<double>(3.881*sr, 0.0);
	del_b = new delay<double>(3.903*sr, 0.0);
	del_c = new delay<double>(7.721*sr, 0.0);
	del_d = new delay<double>(7.654*sr, 0.0);

	bop = new bin_op(NLmix);
	bup = new bin_op(Mult);

	k = 0;
	m = 0;
	xd = yd = 0;
	x1 = y1 = 0;
	fbx = fby = 0;
	sa = 0;
	t = 0;
}



merge::~merge()
{
	del_a = 0;
	del_b = 0;
	del_c = 0;
	del_d = 0;
	bop = 0;
	bup = 0;
	delete del_a; delete del_b;
	delete del_c; delete del_d;
	delete bop; delete bup;
}

void merge::combine(const double *ibuf, const double *dbuf, double *obuf, int sz)
{

	for(int i=0; i<2*sz; i += 2)
		{
		x = ibuf[i];
		y = ibuf[i+1];
		u = dbuf[i];
		v = dbuf[i+1];
		x1 = xd;
		y1 = yd;
		xd = .5*(del_a->z(.7*x - .3*fby) + x1);
		yd = .5*(del_b->z(.7*y - .3*fbx) + y1);

		A = tr.detect(x+y);
		sa += (A > 0) ? A : 0;
		if(sa > 15)
			regime_change();

		x = .8*x + .2*xd;
		y = .8*y + .2*yd;

		L = bop->process(x, u);
		R = bup->process(y, v);

		fbx = .8*del_c->z(L) + .2*fbx;
		fby = .8*del_d->z(R) + .2*fby;
		obuf[i  ] = L;
		obuf[i+1] = R;
		++t;
		}
}


void merge::regime_change()
{
	sa = 0;
	if(++k > 1100)
		{
		bop->next_method();
		k = 35*L + 65*R;
		printf("bop @ %.2f sec, k: %d \n", (double)t/sr, k);
		}

	if(++m >= 1200)
		{
		bup->next_method();
		m = -k;
		printf("bup @ %.2f sec, m = %d \n", (double)t/sr, m);
		}

}



int main()
{

	SNDFILE *inf, *ind, *ouf;
	SF_INFO head, ohead;

	if( !(inf = sf_open("batir.wav", SFM_READ, &head)) )
		{
		printf("can't find the file batir.wav \n\n");
		exit(1);
		}
	ind = sf_open("batir.wav", SFM_READ, &head);	// read again, delayed
	ohead = head;
	ohead.frames = 0;
	ohead.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	ouf = sf_open("demolir_et_demolir.wav", SFM_WRITE, &ohead);

	const int L = 1 << 7;
	const int ch = head.channels;
	const int sr = head.samplerate;
	const int N = ch*L;

	//double ibuf[N], dbuf[N], obuf[N]; doesn't work in all compilers !
    double *ibuf = new double[N+1];
    double *dbuf = new double[N+1];
    double *obuf = new double[N+1];

    /*//malloc
    double * ibuf = (double*)malloc((N+1)*sizeof(double));
	double  * dbuf = (double*)malloc((N+1)*sizeof(double));
	double * obuf = (double*)malloc((N+1)*sizeof(double));

	if( ibuf == NULL || dbuf == NULL || obuf == NULL)
    {
		printf("Memory alloc error !\n\n");
		exit(0);
    }*/


	sf_count_t numFrames = head.frames;
	double sec = (double) numFrames / sr;
	long int read_at = sf_seek(ind, numFrames/3, SEEK_SET);	// samples from start
	printf("length in: %.2f sec = %ld frames. 2nd.point: %ld \n", sec, numFrames, read_at);


	merge f_processor(sr);
	int items_read = 0;

	while(sf_read_double(inf, ibuf, N))
		{
		if( ( items_read = sf_read_double(ind, dbuf, N) ) < N)
			{
			sf_seek(ind, 0, SEEK_SET);	// back to start
			sf_read_double(ind, (dbuf + items_read), N-items_read);
			}

		f_processor.combine(ibuf, dbuf, obuf, L);
		sf_write_double(ouf, obuf, N);
		}


    //malloc free
    /*
    free(ibuf);
    free(dbuf);
    free(obuf);
    */

	sf_close(inf);
	sf_close(ind);
	sf_close(ouf);

    ibuf = 0;
    delete ibuf;
    dbuf = 0;
    delete dbuf;
    obuf = 0;
    delete obuf;

	return 0;
}
