/*
	statement.cpp
	Copyright Risto Holopainen 2021

	Generator for the composition Statement
	wavesegment generator

	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef uint
#define uint unsigned int
#endif

#include "../util/sequencer.h"
#include <cstdio>
#include <sndfile.h>


typedef struct
{
	unsigned int T;
	double x;
} segm;



class lop
{
	public:
		lop();
		double filter(double);

	private:
		double x1, x2, x3, x4, y;
		double h0, h1, h2, h3, h4;
		double g, b;
};

lop::lop()
{
	x1 = x2 = x3 = x4 = y = 0.0;
	b = 0.5;
	g = 1-b;

	h0 = g * 1/15.;
	h1 = g * 2/15.;
	h2 = g * 1/5.;
	h3 = g * 4/15.;
	h4 = g * 1/3.;

}

double lop::filter(double x0)
{

	y *= b;
	y += h4*x4 + h3*x3 + h2*x2 + h1*x1 + h0*x0;

	x4 = x3;
	x3 = x2;
	x2 = x1;
	x1 = x0;

	return y;
}


uint asphalt(sequencer<segm> &S, double *road, uint maxlength)
{
	segm H = S.tick();
	double u = 0;
	static double y = 0;
	double loss = 0.99985;
	static lop LP;
	uint t = 0;

	while( t<H.T && t<maxlength )
		{
		y += (y > -0.9 && y < 0.9) ? .05*H.x : 0;
		y *= loss;
		u = LP.filter( .5 * (y + H.x) );
		road[t] = u;
		++t;
		}

	return t;
}


void init_road(sequencer<segm> &G)
{

	segm ss = {1,0.0};
	int N = 45;		// initial number of segments -- 24
	deque <segm> DS;

	for(int i=0; i<N; i++)
		{
		DS.push_back(ss);
		ss.T = 15 + (i%7)*(i%13);
		ss.x = (i&1) - 0.5;
		ss.x *= (i%11) * 0.125;
		}
	DS.push_back(ss);
	G.replace(DS);
}


void start_here(deque<segm> &D)
{
	segm s = {1, 0.0};
	D.push_back(s);
}


void printsegm(const deque<segm> &ss)
{
	const int N = ss.size();

	for(int i=0; i<N; i++)
		printf("%2d: \t%.2d, \t %7.3f \n", i, ss[i].T, ss[i].x);
}




class trigger
{
	public:
		trigger();
		~trigger();
		bool run();
		void mod();

	private:
		deque <uint> delta_t;
		sequencer <uint> *Timer;
		bool trig;
		uint t;
		uint dt;
		uint k;
};



trigger::trigger()
{
	delta_t.assign(3, 1);	// 3,7
	Timer = new sequencer <uint> (delta_t);
	trig = false;
	t = 0;
	dt = 0;
	k = 1;
}


trigger::~trigger()
{
	Timer = 0;
	delete Timer;
}


bool trigger::run()
{
	trig = false;
	if(++t >= dt)
		{
		t = 0;
		trig = true;
		dt = Timer->tick();
		}
	return trig;
}


void trigger::mod()
{
	uint s = Timer->read(0);
	if (s < 63)
		{
		++s;
		Timer->write(0, s);
		}
	if (Timer->length() < 12)
		Timer->add_left(k);
	else
		Timer->slice(3, 7);

	k = 11 + k%121;
}



void roadwork(sequencer<segm> &rd)
{
	segm ss;
	uint N = rd.length();
	static int t = 0, incr = 1;
	++t;
	if(N==24) t = 0;
	int u = t%5;
	int a = t%7;
	int b = t%9;

	switch (u)
		{
		case 0 :
			{
			rd.reverse();
			break;
			}
		case 1 :
			{
			rd.stride(1 + a);
			break;
			}
		case 2 :
			{
			rd.swap(a, b);
			break;
			}
		case 3 :
			{
			if(N > 6)
				{
				int M = N/2;
				ss = rd.read(M);
				rd.write(M-2, ss);
				rd.write(M+2, ss);
				}
			rd.stride(1);	// resetting
			break;
			}
		default :
			{
			ss = rd.tick();
			ss.T += incr;
			if(ss.T > 200)
				incr = -2;
			else if(ss.T < 5)
				incr = 1;

			double z = ss.x;
			ss.x = 1 - 2*z*z;
			if(z > 0)
				ss.x *= -1;
			rd.write(1, ss);
			ss = rd.read(N-1);
			ss.x *= 0.75;
			rd.write(N-1, ss);
			}
		}
}



int road_remap(sequencer <segm> &Roadie, segm &u, int k)
{
	static int m = 0;
	int N = Roadie.length();
	if(u.T < 200) u.T *= 1.8;
	if(k%59 == 0 && N < 60)
		N = Roadie.add_left(u);

	if(k%61 == 1 && N < 60)
		N = Roadie.add_right(u);

	if(k%177 == 2)
		{
		N = Roadie.length();
		if(N > 16)
			Roadie.slice(2, N-2);

		++m %= N-5;
		u = Roadie.read(m);
		if( u.x < 0.05 && u.x > -0.06 )
			u.x *= 3.0;
		if(u.T < 250)
			u.T += 1 + 3 * (k%8);	// make segm longer
		else
			u.T /= 50;

		Roadie.write(m, u);
		u.x = - 0.7*u.x;
		Roadie.add_right(u);
		}

	if(Roadie.length() > 56)
		Roadie.slice(5, 15);

	const int K = 21;
	if( k%14 == 3 && Roadie.length() == K )
		{
		printf("ordering: road @ k=%d\n", k);
		// put in order.strict
		segm v;
		int s = 0;
		double y=0;
		for(int i=0; i<K; i++)
			{
			v = Roadie.read(i);
			s += v.T;
			}
		v.T = s/K;
		if (v.T > 100) v.T *= 0.8;
		for(int i=0; i<K; i++)
			{
			y = (i&1) ? 1 : -1;
			v.x = .9*y + .1*v.x;
			Roadie.write(i, v);
			}
		}

	return Roadie.length();
}


/*global*/
int a_count = 0;
int g1_count = 0;


double caput(const segm &s)
{
	// amplitude controller
	static lop lupus, loplop;
	static double A = 0;
	double g;
	double eps = 1E-05;

	A += eps;
	if( (s.x > 0.85 || s.x < - 0.80) && s.T > 25 && s.T < 40)
		{
		A *= 0.02;
		++a_count;
		}

	if (A > 1.0)
		{
		A -= 5*eps;
		++g1_count;
		}

	g = lupus.filter(A);
	return loplop.filter(g);
}




uint road_works(deque<segm> &S, double *road, uint length)
{
// quasi delay -- echo
// make length of road segment

	static int k = 0;
	segm H = S[k];
	double u = 0;
	static lop LP;
	static uint t = 0;
	uint n = 0;

	while( n<length )
		{
		u = LP.filter(H.x);
		road[n] = u;
		if(++t >= H.T)
			{
			++k %= S.size();
			H = S[k];
			S[k].x *= 0.995;
			t = 0;
			if(k == 1 && S.size() > 5)
				S.pop_front();
			}
		++n;
		}

	return t;
}


uint roadsplit(sequencer<segm> &S, deque<segm> &D)
{
	D = S.copy();
	const int N = S.length();
	if(N < 6)
		return N;

	const int M = N/2;
	S.slice(1, N-1);
	D.erase(D.begin(), D.end()-M);

	segm u;
	for(uint i=0; i<S.length(); i++)
		{
		u = S.read(i);
		u.x *= 0.85;
		if(u.T > 60) u.T -= 5;
		S.write(i, u);
		}

	return D.size();
}

double inspect(const sequencer<segm> &R, uint &maxsegl)
{
	// avg T
	segm u;
	maxsegl = 0;
	uint s = 0;
	const int N = R.length();

	for(int i=0; i<N; i++)
		{
		u = R.read(i);
		if (u.T > maxsegl) maxsegl = u.T;
		s += u.T;
		}
	return (double) s / N;
}


int main()
{
	segm u = {1, 0.1};
	deque<segm> st;
	start_here(st);
	sequencer <segm> Roadie(st);
	init_road(Roadie);

	deque<segm> echo;
	start_here(echo);

	uint seglen=0;
	const uint maxseg = 600;
	double xx[maxseg];
	double yy[maxseg];	// for echo
	trigger Cosmic;

	const uint L = maxseg;
	double buf[2*L];
	double amp;

	SNDFILE *ouf;
	SF_INFO head;
	head.samplerate = 96000;
	head.channels = 2;
	head.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	ouf = sf_open("statement.wav", SFM_WRITE, &head);

	const int stop_sign = head.samplerate * 60 * 5.5;
	bool stop_now = false;

	int k=0;
	for(int N=0, tcount=0, splitter_counter=0; k<15000 && !stop_now; k++)
		{
		N = Roadie.length();
		for(int i=0; i<N; i++)
			{
			seglen = asphalt(Roadie, xx, maxseg);
			u = Roadie.tick();
			road_works(echo, yy, seglen);
			for(uint j=0, n=0; j<seglen && j<L; j++, tcount++, n += 2)
				{
				amp = caput(u);
				buf[n  ] = amp*xx[j];
				buf[n+1] = yy[j];

				if(tcount > stop_sign)
					stop_now = true;
				}
			sf_write_double(ouf, buf, 2*seglen);
			}

		if ( Cosmic.run() )
			roadwork(Roadie);

		N = road_remap(Roadie, u, k);

		if (N > 24 && N < 37)
			{
			uint longest_s=0;
			double avgT = inspect(Roadie, longest_s);
			++splitter_counter;
			if(k % 75 == 5 || avgT > 100 || longest_s > 200 || splitter_counter == 1000)
				{
				N = roadsplit(Roadie, echo);
				printf("k%4d -> \t avg T %.3f, max %2d , N=%d splcnt: %2d\n",
					k, avgT, longest_s, N, splitter_counter);
				splitter_counter = 0;
				}
			}
		}

	Roadie.copy(st);
	printsegm(st);

	printf("a-count: %d \t g>1: %d k: %4d\n\n", a_count, g1_count, k);

	sf_close(ouf);
	return 0;
}
