/*
	parleur 
	Copyright Risto Holopainen 2021

	help functions, classes
	for the final countdown
	
	 This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "parloop.h"
#include "../util/envelope.h"
#include "../util/timedomain.h"
#include "../util/biquad2.h"



struct features
{
	double amp;
	double amp1k;
	double zc;
	double var;
};

class analyzer
{
	public:
		analyzer(int sr);
		~analyzer();
		features operator()(double);
		
	private:
		int sr;
		zcr *zc;
		tv *telly;
		halfwave *hw0, *hw1;
		bandpass *bp;
		features phi;
};



analyzer::analyzer(int fs)
{
	sr = fs;
	zc = new zcr(sr / 25);
	telly = new tv(sr / 200);
	hw0 = new halfwave(sr / 15);
	hw1 = new halfwave(sr / 50);
	bp = new bandpass(sr, 1000, 100);
	
}

analyzer::~analyzer()
{
	bp=0;
	hw0=0;
	hw1=0;
	telly=0;
	zc=0;
	delete bp;
	delete hw0;
	delete hw1;
	delete telly;
	delete zc;
}

features analyzer::operator()(double x)
{
	double k1 = bp->filter(x);
	phi.amp1k = hw1->amp(k1);
	phi.amp = hw0->amp(x);
	phi.zc = zc->robust(x);
	phi.var = telly->local(x);

	return phi;
}



class transient
{
	public:
		transient(int sr=48000);
		bool detect(const features &X);
	
	private:
		bool ch;
		double Tr;
		int t;
		double eps;
		double smpt;
		int k;
		double step[3];
};


transient::transient(int sr)
{
	ch = false;
	Tr = 0.5;
	t = 0;
	eps = 5E-07;
	smpt = 1.0 / sr;
	step[0] = .2; step[1] = 1.0; step[2] = .5;
	k = 0;
}

bool transient::detect(const features &X)
{
	ch = false;

	if(X.var > Tr)
		{
		printf("t %4.1f : var %.3f, Tr %.2f \n", smpt*t, X.var, Tr);
		ch = true;
		Tr += step[k];
		if(++k >= 3)
			k = 0;
		}
	++t;
	Tr -= eps;
	return ch;
}




class compenv
{
	public:
		compenv(int sr);
		~compenv();
		double env();
		
	
	private:
		int sr;
		envelope *eg1, *eg2, *eg3;
		looper *lup;
		double A, u1, u2, u3, v;
		boxcar *ECM;
};


compenv::compenv(int fs)
{
	sr = fs;
	eg1 = new envelope(sr, SQUARE, 3, 0.25);
	eg2 = new envelope(sr, RAMP, 4, 0.25);
	eg3 = new envelope(sr, EXPO, 5, 0.25);
	int TL = sqrt(2.)*sr;
	lup = new looper(TL);
	lup->set_loop(sr/5, sr);
	ECM = new boxcar(15);
	A = u1 = u2 = u3 = v = 0;
}

compenv::~compenv()
{
	eg1 = eg2 = eg3 = 0;
	delete eg1; delete eg2; delete eg3;
	lup = 0;
	delete lup;
	ECM = 0;
	delete ECM;
}

double compenv::env()
{
	u1 = eg1->env();
	u2 = eg2->env();
	u3 = eg3->env();
	lup->overwrite(u2 + u3);
	v = .25*lup->read() + .75*v;
	A = tanh(u1 + u2 + u3 + v + .7*A);
	A = ECM->filter(A);
	return A;
}

