
/*
	sequencer
	Copyright Risto Holopainen 2021
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/


#ifndef SEQUENCER_H
#define SEQUENCER_H


#include <deque>

using std::deque;


template <typename T>
class sequencer
{
	public:
		sequencer(const deque<T> &ls);
		T tick();
		void reverse();	// switch reading direction forwards/backwards
		void reset();
		void stride(unsigned int n=1);	// number of steps to move at tick()
		void write(int pos, const T &val);
		T read(int pos) const;
		void swap(int pos_a, int pos_b);
		void replace(const deque<T> &new_list);	// ditch x, replace with new list.
		unsigned int join(const deque<T> &left, const deque<T> &right);	// return new length
		unsigned int add_left(const T &first);	// insert new first element
		unsigned int add_right(const T &last);
		unsigned int slice(int begin, int end);		// cut out start&end
		unsigned int length() const { return N; };
		
		void copy(deque<T> &u) const;
		deque<T> copy() const { return this->x; };

	private:
		deque<T> x;
		int N;
		int n;
		int steps;
		bool forward;
		T first_resort;
};

template <typename T>
sequencer<T>::sequencer(const deque<T> &d)
{
	x = d;
	N = x.size();
	n = 0;
	steps = 1;
	forward = true;
	first_resort = x[0];
}





template <typename T>
T sequencer<T>::tick()
{
	if(N == 0)
		return first_resort;

	T c = x[n];
	if(forward)
		{
		n += steps;
		while(n >= N)
			n -= N;
		}
	else
		{
		n -= steps;
		while(n < 0)
			n += N;
		}

	return c;
}


template <typename T>
void sequencer<T>::reverse()
{
	forward = !forward;
}


template <typename T>
void sequencer<T>::reset()
{
	if(forward)
		n = 0;
	else
		n = N-1;
}


template <typename T>
void sequencer<T>::stride(unsigned int st)
{
	steps = st;
}


template <typename T>
void sequencer<T>::swap(int a, int b)
{
	if( a < N && b < N && a >= 0 && b >= 0)
		{
		T tmp = x[a];
		x[a] = x[b];
		x[b] = tmp;
		}
}


template <typename T>
void sequencer<T>::write(int pos, const T &val)
{
	if(pos<N && pos>=0)
		x[pos] = val;
	else
		return;
}


template <typename T>
T sequencer<T>::read(int pos) const
{
	T u;
	if(pos<N && pos>=0)
		u = x[pos];
	return u;
}



template <typename T>
void sequencer<T>::replace(const deque<T> &y)
{
	x = y;
	N = x.size();
}


template <typename T>
unsigned int sequencer<T>::join(const deque<T> &first, const deque<T> &second)
{
	x = first;
	x.insert(x.end(), second.begin(), second.end());
	N = x.size();
	return N;
}

template <typename T>
unsigned int sequencer<T>::add_left(const T &first)
{	
	// insert new first element
	x.push_front(first);
	N = x.size();
	return N;
}

template <typename T>
unsigned int sequencer<T>::add_right(const T &last)
{	
	// insert new last element
	x.push_back(last);
	N = x.size();
	return N;
}


template <typename T>
unsigned int sequencer<T>::slice(int begin, int end)
{
	if(end < begin)
		return -1;

	int i=0, j=x.size();
	while(i++ < begin)
		x.pop_front();

	while(j-- > end)
		x.pop_back();

	N = x.size();
	return N;
}

template <typename T>
void sequencer<T>::copy(deque<T> &u) const
{
	u = x;
}

#endif // def. sequencer.h
